#include "motif-in.h"
#include "config.h"
#include "momo-output.h"
#include "momo-html-string.h"
#include "io.h"
#include "motif-spec.h"
#include "seq-reader-from-fasta.h"
#include "ceqlogo.h"
#include "momo-algorithm.h"
#include "momo-modl.h"
#include "fisher_exact.h"

const int MAX_HTML_MATCHES = 1000;

/**********************************************************************
 * This function saves MOMO results as a MEME motif file.
 *********************************************************************/
void print_momo_text_file(FILE *html_file, MOMO_OPTIONS_T options, SUMMARY_T summary) {
  
  fprintf(html_file, "MEME version %s\n\n", VERSION);
  fprintf(html_file, "Alphabet= %.*s\n\n", 20, summary.alph->symbols + 1);
  fprintf(html_file, "Background letter frequencies\n");
  
  const char* alph_letters = summary.alph_letters;
  ARRAY_T * bg_freqs = summary.bg_freqs;
  int i, j, k, l;
  
  for (i = 0; i < strlen(alph_letters); ++i) {
    fprintf(html_file, "%c %8.6f ", alph_letters[i], get_array_item_defcheck(i, bg_freqs));
  }
  fprintf(html_file, "\n\n");

  ARRAYLST_T * mod_table_keys = summary.mod_table_keys;
  
  for (i = 0; i < arraylst_size(mod_table_keys); i++) {
    HASH_TABLE_ENTRY * hash_entry = arraylst_get(i, mod_table_keys);
    MOD_INFO_T * mod_entry = (MOD_INFO_T *) hash_get_entry_value(hash_entry);
    ARRAYLST_T* motifs = mod_entry->motifinfos;
    for (j = 0; j < arraylst_size(motifs); ++j) {
      MOTIF_INFO_T* motifinfo = arraylst_get(j, motifs);
      MOTIF_T* motif = motifinfo->motif;

      fprintf(html_file, "MOTIF %s\n", get_motif_id(motif));
      unsigned long motif_list_size = (unsigned long) arraylst_size(motifinfo->seqs);
      double m1, e1;
      exp10_logx(motif->log_evalue/log(10.0), m1, e1, 1);
      fprintf(html_file, "letter-probability matrix: alength= %d w= %d nsites= %lu E= %3.1fe%+04.0f\n", (int) (strlen(alph_letters)), options.width, motif_list_size, m1, e1);

      MATRIX_T* freqs = get_motif_freqs(motif);
      for (k = 0; k < options.width; k++) {
	for (l = 0; l < strlen(alph_letters); l++) {
	  fprintf(html_file, "%8.6f\t", get_matrix_cell_defcheck(k, l, freqs));
	}
	fprintf(html_file, "\n");
      }
      fprintf(html_file, "\n");
    }
  }
};

void momo_print_version(FILE *html_file) {
  fprintf(html_file, "MoMo version %s (Release date: %s)", VERSION, ARCHIVE_DATE);
};

void momo_finish_tsv_file(FILE *tsv_file, MOMO_OPTIONS_T options) {
  // Finish the TSV output
  char *version_message = "# MoMo (Modification Motifs): Version " VERSION " compiled on " __DATE__ " at " __TIME__ "\n";
  fprintf(tsv_file, "\n%s", version_message);
  fprintf(tsv_file, "# The format of this file is described at %s/%s.\n", SITE_URL, "doc/momo-output-format.html");
  fprintf(tsv_file, "# %s\n", options.command_line);
}

void momo_print_interpreting(FILE *html_file, MOMO_OPTIONS_T options) {
  fprintf(html_file, 
    "For further information on how to interpret these results please access "
    "<a href=\"%s/doc/momo-output-format.html\">%s/doc/momo-output-format.html</a>.<br />\n"
    "To get a copy of the MoMo software please access "
    "<a href=\"%s\">%s</a>.\n",
    SITE_URL, SITE_URL, SOURCE_URL, SOURCE_URL);
}

void momo_print_citing(FILE *html_file, MOMO_OPTIONS_T options) {
  fprintf(html_file, 
    "If you use MoMo in your research, please cite the following paper:<br>\n"
    "%s\n<a href='%s'>[full text]</a>\n", 
    options.paper, options.link);
};

void momo_print_command_line(FILE *html_file, MOMO_OPTIONS_T options) {
  fputs(options.command_line, html_file);
};

void print_algorithm_name(MOMO_OPTIONS_T options, FILE *fp) {
  ALGORITHM_T algorithm = options.algorithm;
  if (algorithm == Simple) {
    fprintf(fp, "simple");
  } else if (algorithm == Motifx) {
    fprintf(fp, "motif-x");
  } else if (algorithm == Modl) { 
    fprintf(fp, "MoDL");
  } else {
    fprintf(fp, "UNKNOWN");
  }
}

void momo_print_parameters(FILE *html_file, MOMO_OPTIONS_T options) {
  int i;
  
  fprintf(html_file, "PARAMETERS:\n\n");
  
  // Algorithm
  fprintf(html_file, "algorithm: "); print_algorithm_name(options, html_file); fprintf(html_file, "\n");
  
  // PTM files
  ARRAYLST_T* phospho_filenames = options.phospho_filenames;
  fprintf(html_file, "post-translationally modified peptide filenames: \n");
  for (i = 0; i < arraylst_size(phospho_filenames); ++i) {
    char* phospho_filename = arraylst_get(i, phospho_filenames);
    fprintf(html_file, "\tfile %d: %s\n", i+1, phospho_filename);
  }

  // PTM types and mod pep column.
  FILETYPE_T filetype = options.filetype;
  FILETYPE_T bg_filetype = options.bg_filetype;
  if (filetype == Psm) {
    fprintf(html_file, "PTM filetype: %s\n", options.psm_type ? options.psm_type : "PSM");
    fprintf(html_file, "modified peptide column: '%s'\n", options.sequence_column);
  } else if (filetype == Prealigned) {
    fprintf(html_file, "PTM filetype: Raw\n");
  } else { // fasta
    fprintf(html_file, "PTM filetype: FASTA\n");
  }

  // Context sequence
  if (options.protein_database_filename) fprintf(html_file, "protein database filename: %s\n", options.protein_database_filename);
  if (options.protein_database_filename) {
    if (bg_filetype == Psm) {
      fprintf(html_file, "protein database format: PSM (error!)\n");
    } else if (bg_filetype == Prealigned) {
      fprintf(html_file, "protein database format: Raw\n");
    } else { // fasta
      fprintf(html_file, "protein database format: FASTA\n");
    }
  }
  
  fprintf(html_file, "motif width: %d\n", options.width);

  fprintf(html_file, "filter: %s\n", (options.filter ? "true" : "false"));
  if (options.filter) {
    fprintf(html_file, "\tfilter field: '%s'\n", options.filter_field);
    if (options.filter_type == Le) {
      fprintf(html_file, "\tfilter type: <=\n");
    } else if (options.filter_type == Lt) {
      fprintf(html_file, "\tfilter type: <\n");
    } else if (options.filter_type == Eq) {
      fprintf(html_file, "\tfilter type: =\n");
    } else if (options.filter_type == Gt) {
      fprintf(html_file, "\tfilter type: >\n");
    } else if (options.filter_type == Ge) {
      fprintf(html_file, "\tfilter type: >=\n");
    } else {
      fprintf(html_file, "\tfilter type is unknown!\n");
    }
    fprintf(html_file, "\tfilter threshold: %g\n", options.filter_threshold);
  }

  fprintf(html_file, "remove unknowns: %s\n", (options.remove_unknowns ? "true" : "false"));
  fprintf(html_file, "eliminate repeats: %s\n", options.eliminate_repeat_width > 0 ? "true" : "false");
  if (options.eliminate_repeat_width > 0)
    fprintf(html_file, "\teliminate repeat width: %d\n", options.eliminate_repeat_width);
  fprintf(html_file, "min occurrences: %d\n", options.min_occurrences);
  fprintf(html_file, "single motif per mass: %s\n", (options.single_motif_per_mass ? "true" : "false"));
  fprintf(html_file, "hash fasta: %s\n", (options.hash_fasta ? "true" : "false"));
  fprintf(html_file, "\thash fasta width: %d\n", options.hash_fasta_width);

  if (options.algorithm == Motifx) {
    fprintf(html_file, "score threshold: %g\n", options.score_threshold);
    fprintf(html_file, "binomial p-value calculations: %s\n", options.harvard ? "inaccurate (emulate original motif-x)" : "accurate");
  }

  if (options.algorithm == Modl) {
    fprintf(html_file, "max motifs: %d\n", options.max_motifs);
    fprintf(html_file, "max iterations: %d\n", options.max_iterations);
    fprintf(html_file, "max no decrease iterations: %d\n", options.max_no_decrease);
  }
  
  // ints

  fprintf(html_file, "\n");
};

void momo_print_peptides(
  FILE *html_file, 
  MOMO_OPTIONS_T options, 
  SUMMARY_T summary
) {
  fprintf(html_file, "    <ul>\n");
  fprintf(html_file, "      <li>Algorithm: "); print_algorithm_name(options, html_file); fprintf(html_file, "</li>\n");
  fprintf(html_file, "      <li>Number of Mods: %lu</li>\n", summary.num_mod);
  fprintf(html_file, "      <li>Number of Mod Types: %lu</li>\n", summary.num_modtype);
  fprintf(html_file, "      <li>Number of Mods Passing Filters: %lu</li>\n", summary.num_mod_passing);
  fprintf(html_file, "      <li>Number of Mod Types Passing Filters: %lu</li>\n", summary.num_modtype_passing);
  fprintf(html_file, "    </ul><br>\n");
} // momo_print_peptides
  
void momo_print_motifs(
  FILE *html_file, 
  FILE *tsv_file, 
  MOMO_OPTIONS_T options, 
  SUMMARY_T summary
) {
  ARRAYLST_T * mod_table_keys = summary.mod_table_keys;
  int i, j, k, m;
  bool pvalue_accurate = (!options.db_background && options.algorithm == Motifx);

  for (i = 0; i < arraylst_size(mod_table_keys); ++i) {
    HASH_TABLE_ENTRY * hash_entry = arraylst_get(i, mod_table_keys);
    MOD_INFO_T * mod_entry = (MOD_INFO_T *) hash_get_entry_value(hash_entry);
    ARRAYLST_T * motifs = mod_entry->motifinfos;

    // Skip mods with no motifs.
    if (arraylst_size(motifs) == 0) continue;

    for (j = 0; j < arraylst_size(motifs); ++j) {
      MOTIF_INFO_T* currmotifinfo = arraylst_get(j, motifs);
      MOTIF_T * currmotif = currmotifinfo->motif;
      char* motifid = currmotif->id + 1;

      char* motif_name = mm_malloc(strlen(motifid) + 2);
      // Add a central 'X' in the motif name if doing single motif per mass.
      for (k=m=0; k < strlen(motifid); k++) {
        // Check for single motif per mass (may be "__" or "_number").
        if ( (k == 0) && (motifid[0]-'0' >= 0 && motifid[0]-'0' <= 9) ) {
          // Add missing central 'X' to name.
          motif_name[m++] = 'X';
          motif_name[m++] = motifid[0];
	} else if (motifid[k] == '_' && (motifid[k+1] == '_' || (motifid[k+1]-'0' >= 0 && motifid[k+1]-'0' <= 9))) { 
          // Convert '_' to 'X' for (missing) central residue.
          motif_name[m++] = '_';
          motif_name[m++] = 'X';
	} else {
	  motif_name[m++] = motifid[k];
	}
      }
      motif_name[m] = '\0';

      char* motif_png = mm_malloc(strlen(motifid) + 5);
      motif_png[0] = '\0';
      strncat(motif_png, motifid, strlen(motifid));
      strncat(motif_png, ".png", 4);
      
      char* logo_file = mm_malloc(strlen(options.output_dirname) + strlen(motifid) + 2);
      logo_file[0] = '\0';
      strncat(logo_file, options.output_dirname, strlen(options.output_dirname));
      strncat(logo_file, "/", 1);
      strncat(logo_file, motifid, strlen(motifid));
      CL_create1(currmotif, false, false, "MoMo", logo_file, false, true);
      
      if (options.algorithm == Motifx || options.algorithm == Modl) {
        double score = currmotifinfo->score;
        int fg_matches, fg_size, bg_matches, bg_size;
        // To emulate original motif-x we use the counts in the remaining sequences;
	// by default we use all the matches in all the sequences.
        if (options.algorithm == Motifx && options.harvard) {
          fg_matches = currmotifinfo->fg_matches;
          fg_size = currmotifinfo->fg_size;
          bg_matches = currmotifinfo->bg_matches;
          bg_size = currmotifinfo->bg_size;
        } else {
	  fg_matches = currmotifinfo->afg_matches;
	  fg_size = currmotifinfo->afg_size;
	  bg_matches = currmotifinfo->abg_matches;
	  bg_size = currmotifinfo->abg_size;
        }
        double bg_fold = (double) fg_matches*bg_size / (bg_matches*fg_size);

        // Get p-value, adjusted p-value and E-value.
        double n_tests = currmotifinfo->n_tests;
	double m1, m2, m3;
	double e1, e2, e3;
        double log_pvalue=0, log_norm_pvalue=0, log_evalue=0;

        // Calculate significance if motif-x.
	log_pvalue = getLogFETPvalue(fg_matches, fg_size, bg_matches, bg_size, false);
	exp10_logx(log_pvalue/log(10.0), m1, e1, 1);
        if (pvalue_accurate) {
          double log_n_tests = log(n_tests);
	  log_norm_pvalue = LOGEV(log_n_tests, log_pvalue);
	  exp10_logx(log_norm_pvalue/log(10.0), m2, e2, 1);
	  log_evalue = log_pvalue + log_n_tests;
	  exp10_logx(log_evalue/log(10.0), m3, e3, 1);
        }
        currmotif->log_evalue = log_evalue;

        // Print HTML 
	fprintf(html_file, 
          "  <li><b>motif:</b> %s  <b>score:</b> %.2f  <b>fg:</b> %d/%d  <b>bg:</b> %d/%d  <b>fg/bg:</b> %.1f  <b>unadjusted <i>p</i>-value:</b> %3.1fe%+04.0f",
          motif_name, score, fg_matches, fg_size, bg_matches, bg_size, bg_fold, m1, e1
        );
        if (pvalue_accurate) { 
	  //fprintf(html_file, "  <b>tests:</b> %0.f  <b>adjusted <i>p</i>-value:</b> %3.1fe%+04.0f  <b><i>E</i>-value:</b> %3.1fe%+04.0f <br>\n", n_tests, m2, e2, m3, e3);
	  fprintf(html_file, "  <b>tests:</b> %0.f  <b>adjusted <i>p</i>-value:</b> %3.1fe%+04.0f <br>\n", n_tests, m2, e2);
        } else {
	  fprintf(html_file, "<br>\n");
        }

        // Print TSV header if first motif.
	if (i==0 && j==0) fprintf(tsv_file, 
          "motif\tscore\tfg match\tfg size\tbg match\tbg size\tfg/bg"
          "\tunadjusted p-value%s",
          //(pvalue_accurate ? "\ttests\tadjusted p-value\tE-value\n" : "\n"));
          (pvalue_accurate ? "\ttests\tadjusted p-value\n" : "\n"));
        // Print TSV.
	fprintf(tsv_file, 
          "%s\t%.2f\t%d\t%d\t%d\t%d\t%.1f\t%3.1fe%+04.0f",
          motif_name, score, fg_matches, fg_size, bg_matches, bg_size, bg_fold, m1, e1
        );
        if (pvalue_accurate) { 
	  //fprintf(tsv_file, "\t%.0f\t%3.1fe%+04.0f\t%3.1fe%+04.0f\n", n_tests, m2, e2, m3, e3);
	  fprintf(tsv_file, "\t%.0f\t%3.1fe%+04.0f\n", n_tests, m2, e2);
        } else {
	  fprintf(tsv_file, "\n");
        }

        // Print the p-values for debugging.
	if (options.printp && options.algorithm == Motifx) { 
          printf("%s score: %.2f ntests: %.0f p-value: %4.2fe%+04.0f norm_p-value: %4.2fe%+04.0f E-value: %4.2fe%+04.0f p-value_accurate? %s\n", 
            motif_name, score, n_tests, m1, e1, m2, e2, m3, e3, pvalue_accurate ? "yes" : "no");
        }
      } else if (options.algorithm == Simple) {
        currmotif->log_evalue = 0;
        // Print HTML 
        fprintf(html_file, "  <li><b>motif:</b> %s <b>fg_matches:</b> %d<br>\n", motif_name, currmotifinfo->fg_matches);
        // Print TSV header if first motif.
	if (i==0 && j==0) fprintf(tsv_file, "motif\tfg_matches\n");
        // Print TSV.
        fprintf(tsv_file, "%s\t%d\n", motif_name, currmotifinfo->fg_matches);
      }
      
      fprintf(html_file, "      <img src=\"%s\" alt=\"sequence logo of motif\"><br>\n", motif_png);
      fprintf(html_file, 
        "      <button onclick=\"change_display('occ_%d_%d')\">Show/Hide Motif Occurrences</button>\n"
        "      <div style='display:none' id='occ_%d_%d'>\n"
        "        <pre class='console'>\n",
        i, j, i, j);
      for (k = 0; k < arraylst_size(currmotifinfo->seqs); ++k) {
        char* curr_motifinfo_seq = arraylst_get(k, currmotifinfo->seqs);
        fprintf(html_file, "%s\n", curr_motifinfo_seq);
      }
      fprintf(html_file, 
        "      </pre>\n"
        "    </div>\n"
        "    <br><br>\n"
        "  </li>\n"
      );
      
      // cleanup
      myfree(motif_name);
      myfree(motif_png);
      myfree(logo_file);
    }
    // Print out MoDL log
    if (options.algorithm == Modl) {
      fprintf(html_file,
        "  <li>\n"
        "    <button onclick=\"change_display('log_%d')\"><b>Show/Hide MoDL Log:</b></button>\n"
        "    <div style='display:none' id='log_%d'>\n"
        "      <ul>\n"
        "        <li><b>MoDL Log</b></li>\n",
        i, i);
      MATRIX_T* bg_freqs = NULL;
      bg_freqs = get_count_matrix(bg_freqs, mod_entry->bg_seq_list, NULL, &options, &summary);
      for (j = 0; j < options.width; ++j) {
        for (k = 0; k < strlen(summary.alph_letters); ++k) {
          set_matrix_cell_defcheck(j,k,get_matrix_cell_defcheck(j,k,bg_freqs)/arraylst_size(mod_entry->bg_seq_list),bg_freqs);
        }
      }
      
      ARRAYLST_T* modl_ops = mod_entry->modl_ops;
      ARRAYLST_T* temp_list = arraylst_create();
      double minDL = INFINITY;
      double initDL = 0;
      for (j = 0; modl_ops && (j < arraylst_size(modl_ops)); ++j) {
        MODL_STEP_T* step = arraylst_get(j, modl_ops);
        fprintf(html_file, "      <li>STEP: %d, DL: %.4f <br> \n", j, step->score);
        do_step(step, temp_list, bg_freqs, options.max_motifs, &options, &summary, mod_entry);
        if (j == 0 ) initDL = step->score;
        if (step->score < minDL) minDL = step->score;
        for (k = 0; k < arraylst_size(temp_list); ++k) {
          fprintf(html_file, "        %s<br>\n", regexmotif_to_string(arraylst_get(k, temp_list), mod_entry, &summary, &options));
        }
        fprintf(html_file, "      </li>\n");
      }
      fprintf(html_file, 
        "      </ul>\n"
        "    </div><br>\n"
        "    <b>Final DL: %.4f ( Decrease: %.2f )</b>\n" 
        "  <br><br></li>\n",
        minDL, minDL != INFINITY ? initDL-minDL : 0);
      free_matrix(bg_freqs);
    } // modl
  } // mods
  
} // momo_print_motifs

/**********************************************************************
 * This function saves MOMO results as an HTML file
 *********************************************************************/
void print_momo_html_and_tsv_files(
  FILE *html_file, 
  FILE *tsv_file,
  MOMO_OPTIONS_T options, 
  SUMMARY_T summary
) {
  const int MAX_TAG_SIZE = 1000;
  int html_string_size = strlen(momo_html_string);
  int i = 0;
  for (i = 0; i < html_string_size; ++i) {
    if (momo_html_string[i] != '@') {
      fputc(momo_html_string[i], html_file);
    }
    else {
      char buffer[MAX_TAG_SIZE];
        ++i;
      int j = 0;
      while (momo_html_string[i] != '@' && j < (MAX_TAG_SIZE - 1)) {
        buffer[j] = momo_html_string[i];
        ++j;
        ++i;
      }
      if (momo_html_string[i] != '@') {
        die("MoMo tag buffer length exceeded\n");
      }
      buffer[j] = '\0';
      if (strcmp("version", buffer) == 0) {
        momo_print_version(html_file);
      }
      else if (strcmp("interpreting", buffer) == 0) {
        momo_print_interpreting(html_file, options);
      }
      else if (strcmp("citing", buffer) == 0) {
        momo_print_citing(html_file, options);
      }
      else if (strcmp("command_line", buffer) == 0) {
        momo_print_command_line(html_file, options);
      }
      else if (strcmp("peptides", buffer) == 0) {
        momo_print_peptides(html_file, options, summary);
      }
      else if (strcmp("motifs", buffer) == 0) {
        momo_print_motifs(html_file, tsv_file, options, summary);
      }
      else if (strcmp("parameters", buffer) == 0) {
        momo_print_parameters(html_file, options);
      }
    }
  }
  momo_finish_tsv_file(tsv_file, options);
}

void create_directory(MOMO_OPTIONS_T options) {
  
  const bool PRINT_WARNINGS = false;
  if (create_output_directory(
                              options.output_dirname,
                              options.allow_clobber,
                              PRINT_WARNINGS
                              )
      ) {
    // Failed to create output directory.
    die("Couldn't create output directory %s.\n", options.output_dirname);
  }
}

/**********************************************************************
 * This function saves the MOMO results as a set of files in a
 * directory:
 *
 *   html_filename will be the name of the HTML output
 *   text_filename will be the name of the plain text output
 *
 * allow_clobber will determine whether or not existing files will
 * be overwritten.
 *********************************************************************/
void print_momo_results(MOMO_OPTIONS_T options, SUMMARY_T summary) {
  
  // Create directory for motifs to have a location
  create_directory(options);
  
  // Print HTML and TSV (must come first so it can set the E-values)
  FILE *html_file = fopen(options.html_path, "w");
  if (!html_file) {
    die("Couldn't open file %s for output.\n", options.html_path);
  }
  FILE *tsv_file = fopen(options.tsv_path, "w");
  if (!tsv_file) {
    die("Couldn't open file %s for output.\n", options.tsv_path);
  }
  print_momo_html_and_tsv_files(html_file, tsv_file, options, summary);
  fclose(html_file);
  fclose(tsv_file);

  // Print plain text.
  FILE *text_file = fopen(options.text_path, "w");
  if (!text_file) {
    die("Couldn't open file %s for output.\n", options.text_path);
  }
  print_momo_text_file(text_file, options, summary);
  fclose(text_file);
  
}
